/*********************************************************************

 gssklog_afs.c 

Description:

	client utility routiones need to process the AFS tokesn

 
CVS Information:

    $Source: $
    $Date: $
    $Revision: $
    $Author: $

********************************************************************/

static char *rcsid = "$Header: $";

/**********************************************************************
                             Include header files
**********************************************************************/
#ifdef _WIN32
#include <windows.h>
#include <winsock2.h>
#else
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include <sys/time.h>
#include <arpa/inet.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#include <fcntl.h>

#ifndef _WIN32
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include <sys/time.h>
#include <arpa/inet.h>
#endif

#include <gssapi.h>
#include <des.h>
#include <afs/stds.h>
#include <afs/auth.h>
#include <afs/kautils.h>
#include <afs/ptserver.h>

#include "gssklog.h"
#include "gssklog_comm.h"
#include "gssklog_afs.h"

/* extern gss_OID gss_nt_service_name; */

gss_cred_id_t cred_handle;

/*********************************************************/
 
afs_int32 create_request (
  unsigned char * req,
  int  * reqlen, 
  char *cname,
  char *cinst,
  char *crealm,
  char *sname,
  char *sinst,
  char *srealm,
  afs_int32  lifetime)
{   
	unsigned char *request = req;
    int   slen;
	int   newreqlen;

    newreqlen =3 + strlen(cname) + strlen(cinst) + strlen(crealm) 
			+ 3/*nulls*/ +
    		strlen(sname) + strlen(sinst) + strlen(srealm)  
			+ 3/*nulls*/ + sizeof(afs_int32);
    if (newreqlen > *reqlen) return KAANSWERTOOLONG;
	
    *request++ = (unsigned char) KRB_PROT_VERSION;
    *request++ = (unsigned char) AUTH_MSG_KDC_REQUEST;
	*request++ = (unsigned char) 0; /* for future version # */
    putstr (request, cname);
    putstr (request, cinst);
    putstr (request, crealm);
    putstr (request, sname);
    putstr (request, sinst);
    putstr (request, srealm);
    putint (request, lifetime);
	*reqlen = newreqlen;
    return 0;
}

/*********************************************************/
int
process_response(unsigned char * resp, size_t resplen, 
			char * cell,
			char **reason, int dosetpag)
{
	int slen;
	unsigned char * pointer; 
	unsigned char * endpointer;
	unsigned char version, auth_msg_type;
	int   byteOrder;
	int rc;

	afs_int32 afskvno;
	afs_uint32 startTime;
	afs_uint32 endTime;
	afs_uint32 errTime;
	afs_uint32 errcode;
	
	struct ktc_encryptionKey sessionKey;
	struct ktc_principal client;
	struct ktc_principal server;
	struct ktc_principal errclient;
	unsigned char ticket[MAXKTCTICKETLEN];
	int ticketlen;
	

	memset(&client, 0, sizeof(client));
	memset(&server, 0 , sizeof(server));

	pointer= resp;
	endpointer = pointer + resplen;

	version = *pointer++;
	if (version != KRB_PROT_VERSION) {
		*reason = "Bad response version";
		return 1;
	}

	byteOrder = *pointer++;
	auth_msg_type = byteOrder & ~1;
	if (auth_msg_type == AUTH_MSG_ERR_REPLY) {
		*reason = "Received AUTH_MSG_ERR_REPLY";
		getstr(pointer, errclient.name);
		getstr(pointer, errclient.instance);
		getstr(pointer, errclient.cell);
		getint(pointer, errTime);
		getint(pointer, errcode);
		*reason = pointer;

		return 2;

	} else
	if (auth_msg_type == AUTH_MSG_KDC_REPLY) {
		*reason = "Got response";
		ticketlen = sizeof(ticket);
		rc = parse_response(resp, resplen,
				 &startTime, &endTime,
				 &afskvno,
				 &sessionKey,
				 &client,
				 &server,
				 ticket, &ticketlen);
		if (rc != 0) {
			*reason = "Failed to parse response";
			return 4;
		}
		rc = store_afs_token(&client,
			&server,
			afskvno,
			ticket,
			ticketlen,
			&sessionKey,
			startTime,
			endTime,
			dosetpag);
			
		if (rc != 0) {
			*reason = "Failed to store the token";
		}
		return rc;

	} else {
		*reason = "bad msg type";
		return 3;
	}

}

/*********************************************************/
int
parse_response(unsigned char * resp, size_t resplen,
				afs_uint32 * startTime,
				afs_uint32 * endTime,
				afs_int32 * afskvno,
				struct ktc_encryptionKey * sessionKey,
				struct ktc_principal * client, 
				struct ktc_principal * server,
				unsigned char * ticket,
				int * ticketlen)
{
	int slen;
	unsigned char * pointer; 
	unsigned char * endpointer;
	unsigned char version, auth_msg_type;
	int   byteOrder;
	unsigned char life;
	int i;
	int cipherlen;
	
	pointer= resp;
	endpointer = pointer + resplen;

	version = *pointer++;
	if (version != KRB_PROT_VERSION) {
		return 1;
	}

	byteOrder = *pointer++;
	auth_msg_type = byteOrder & ~1;
	if (auth_msg_type != AUTH_MSG_KDC_REPLY) {
		return 2;
	}

	getstr(pointer, client->name);
	getstr(pointer, client->instance);
	getstr(pointer, client->cell);
	getint(pointer, *startTime);

	if (*pointer != 1) {
		return 3;
	}
	pointer++;

	getint(pointer, *endTime);
	*afskvno = (unsigned char )*pointer++; /* was skiped in some code */
	
	{   short w;
    	memcpy (&w, pointer, sizeof(short));
    	cipherlen  = (int) htons(w);
		pointer = pointer + sizeof(short);
	}
	if (cipherlen < 0 || cipherlen > 2*MAXKTCTICKETLEN) {
		return 4;
	}

	/* now pointing at cipher which is not encrypted, 
	 * as we did not need to. */

	if ((pointer + cipherlen) != endpointer) {
		return 5;
	}
	
	memcpy (sessionKey, pointer, (int)sizeof(*sessionKey));
	pointer = pointer + sizeof(*sessionKey);

	getstr(pointer, server->name);
	getstr(pointer, server->instance);
	getstr(pointer, server->cell);
	life = *pointer++;
	i = *pointer++;
	*afskvno = i;
	i = *pointer++;
	if (i > *ticketlen) {
		return 6;
	}
	*ticketlen = i;
	memcpy(ticket, pointer, *ticketlen);

	return 0;
}

/*
 * Store the returned credentials as an AFS "token" for the user
 * "AFS ID <userid>".
 */
int 
store_afs_token(
    struct ktc_principal *client,
	struct ktc_principal *server,
    afs_int32 tkt_type,
    unsigned char *ticket_p,
    int ticket_len,
    struct ktc_encryptionKey * session_key,
    afs_int32 starttime,
    afs_int32 endtime,
    int set_pag)
{
	int rc;
	int status;
    struct ktc_token token;
	afs_int32 viceId;
	char username[BUFSIZ];  /* To hold client username structure */

#ifdef DEBUG
    fprintf(stderr,"afs_store_token\n");
#endif

    token.startTime = starttime;
    token.endTime = endtime;
    memcpy((char *) &token.sessionKey, session_key, sizeof(token.sessionKey));
    token.kvno = (short )tkt_type;
    token.ticketLen = ticket_len;
    if (ticket_len > MAXKTCTICKETLEN) {
	fprintf(stderr, "ticket is too long (length %d). Maximum length accepted by AFS cache manager is %d\n", ticket_len, MAXKTCTICKETLEN);
	 return 5;
    }
    memcpy( (char *) token.ticket, (char*) ticket_p, ticket_len);

	/* 
	 * Use PTS to find ViceID for this user 
	 * Replace the actual principal with this number
	 */

	strcpy (username,client->name);
	if (client->instance[0])  
	{
		strcat (username, ".");
		strcat (username, client->instance);
    }
#ifdef DEBUG
        fprintf(stderr,"username=%s\n",username);
#endif

#ifndef _WIN32
#ifndef AFSDIR_CLIENT_ETC_DIRPATH
#define AFSDIR_CLIENT_ETC_DIRPATH "/usr/vice/etc"
#endif

	if (!pr_Initialize (0, AFSDIR_CLIENT_ETC_DIRPATH, server->cell))
	{
		status = pr_SNameToId (username, &viceId);
		pr_End();
	}

    if ((status == 0) && (viceId != ANONYMOUSID))
	{
        sprintf (client->name, "AFS ID %d", viceId);
    	strcpy(client->instance, "");
	}
#else
	/* On unix the pr_Initialize convers this to lower case!!!*/
	lcstring(server->cell, server->cell, sizeof(server->cell));

#endif

#ifdef DEBUG
    fprintf(stderr,"client->name=%s\n", client->name);
	fprintf(stderr,"client->instance=%s\n", client->instance);
	fprintf(stderr,"client->cell=%s\n",client->cell);
	fprintf(stderr,"server->name=%s\n", server->name);
	fprintf(stderr,"server->instance=%s\n", server->instance);
	fprintf(stderr,"server->cell=%s\n",server->cell);
#endif

#ifndef AFS_SETTOK_SETPAG
#define AFS_SETTOK_SETPAG 1
#endif
    rc =  ktc_SetToken(server, &token, client,
			 set_pag ? AFS_SETTOK_SETPAG : 0);

#ifdef DEBUG
        fprintf(stderr,"ktc_SetToken=%d\n",rc);
#endif
	return rc;
}

/******************************************************************/
afs_int32 ktohl_internal (char flags, afs_int32 l)
{
    if (flags & 1) {
	unsigned char *lp = (unsigned char *)&l;
	afs_int32	       hl;
	hl = *lp + (*(lp+1) << 8) + (*(lp+2) << 16) + (*(lp+3) << 24);
	return hl;
    }
    return ntohl(l);
}
