/**********************************************************************

gssklog_comm.h:

Description:
    This header file used internally by the gssklog routines

CVS Information:

    $Source: $
    $Date: $
    $Revision: $
    $Author: $

**********************************************************************/

#ifndef _gssklog_COMM_H
#define _gssklog_COMM_H
#ifdef  __cplusplus
extern "C" {
#endif

/**********************************************************************
                             Include header files
**********************************************************************/

#include <gssapi.h>
/**********************************************************************
                               Define constants
**********************************************************************/

/**********************************************************************
                               Type definitions
**********************************************************************/

typedef struct auth_method_st AUTH_METHOD;
struct auth_method_st
	{
	int version; /* gss = 0 sspi = 1 */
	OM_uint32 (*auth_acquire_cred)(AUTH_METHOD *,
				OM_uint32 *minor_status,
				void ** vpcred_handle,
				char * username,
				char * password);
	OM_uint32 (*auth_init_sec_context) (AUTH_METHOD *,
				OM_uint32 * minor_status,
				void * vpcred_handle,
				void ** vpcontext_handle,
				int	 peer_socket,
				char * service_princ_name);
	OM_uint32 (*auth_wrap) (AUTH_METHOD * method,
				OM_uint32 *minor_status,
				void * pcontext_handle,
				int conf_req_flag,
				gss_qop_t qop_req,
				gss_buffer_t input_message_buffer,
				int * conf_state ,
				gss_buffer_t output_message_buffer);
	OM_uint32 (*auth_unwrap)(AUTH_METHOD * method,
			    OM_uint32 * minor_status,
				void * pcontext_handle,
				gss_buffer_t input_message_buffer,
				gss_buffer_t output_message_buffer,
				int * conf_state,
				gss_qop_t * qop_state);
    OM_uint32(*auth_display_status)(AUTH_METHOD *,
				char * mess,
				OM_uint32 major_stat,
				OM_uint32 minor_stat);
	OM_uint32(*auth_release_buffer)(AUTH_METHOD *,
				OM_uint32 *, 
				gss_buffer_t);
	OM_uint32(*auth_release_cred)(AUTH_METHOD *,
				OM_uint32 *,
				void ** vpcred_handle);
	OM_uint32(*auth_delete_sec_context)(AUTH_METHOD *,
				OM_uint32 *,
				void ** vpcontext_handle);
	};
	
AUTH_METHOD * gssklog_gss_init_method();
#ifdef _WIN32
AUTH_METHOD *gssklog_sspi_init_method();
#endif

/**********************************************************************
                               Global variables
**********************************************************************/

/**********************************************************************
                               Function prototypes
**********************************************************************/

int create_socket(unsigned int);

int
connect_to_server_sockaddr(struct sockaddr_in * saddr);

int
connect_to_server(char *, unsigned int);

int setup_connection();
int establish_connection( int);

int connection(char *, unsigned int, char *);

int 
send_message(AUTH_METHOD * method,
		 int peer_socket,
	     void * pcontext_handle,
	     gss_buffer_t input_message_buffer,
		 int wrap_flag,
	     int encrypt_flag) ;

int 
receive_message(AUTH_METHOD * method,
	     int peer_socket,
	     void * pcontext_handle,
	     gss_buffer_t output_message_buffer,
		 int wrap_flag);

#endif  /* _gssklog_COMM_H */
