/*
 gssklog_gss.c: 

Description:

	Using the fact that you have a gsi or K5 credential,
	get an AFS token. 
 
CVS Information:

    $Source: $
    $Date: $
    $Revision: $
    $Author: $

********************************************************************/

static char *rcsid = "$Header: $";

/**********************************************************************
                             Include header files
**********************************************************************/

#include <stdio.h>
#include <stdlib.h>

#ifdef _WIN32
#include <windows.h>
#include <winsock2.h>
#define wsclose closesocket
#else
#define wsclose close
#endif

#include <string.h>
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#include <sys/types.h>

#ifndef _WIN32
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include <sys/time.h>
#include <arpa/inet.h>
#endif

#ifdef _WIN32
#define GSS_USE_FUNCTION_POINTERS
#define CALL_GSS(fun) (*FP_##fun)
#else
#define CALL_GSS(fun) fun
#endif
#include "gssklog_comm.h"
#include "gssklog.h"

typedef struct cred_handle_plus_struct {
	gss_cred_id_t cred_handle;
	gss_OID mech_use;
} cred_handle_plus_desc, *cred_handle_plus;

static gss_OID_desc my_nt_service_name_desc =
    {10, (void *)"\x2a\x86\x48\x86\xf7\x12\x01\x02\x01\x04"};
static gss_OID my_nt_service_name = &my_nt_service_name_desc;

static gss_OID_desc gss_mech_krb5_desc = 
	{9, (void*)"\052\206\110\206\367\022\001\002\002"};
static gss_OID gss_mech_krb5 = &gss_mech_krb5_desc;

static gss_OID_set_desc gss_mech_krb5_set_desc =
	{1, (gss_OID) &gss_mech_krb5_desc};

static gss_OID_set gss_mech_krb5_set = &gss_mech_krb5_set_desc;

#ifdef _WIN32
static gss_acquire_cred_type		FP_gss_acquire_cred = NULL;
static gss_delete_sec_context_type  FP_gss_delete_sec_context = NULL;
static gss_display_status_type      FP_gss_display_status = NULL;
static gss_import_name_type			FP_gss_import_name = NULL;
static gss_init_sec_context_type	FP_gss_init_sec_context = NULL;
static gss_release_buffer_type		FP_gss_release_buffer = NULL;
static gss_release_name_type		FP_gss_release_name = NULL;
static gss_wrap_type				FP_gss_wrap = NULL;
static gss_unwrap_type				FP_gss_unwrap = NULL;
static gss_release_cred_type		FP_gss_release_cred = NULL;
static gss_indicate_mechs_type		FP_gss_indicate_mechs = NULL;
static gss_release_oid_set_type		FP_gss_release_oid_set = NULL;
static gss_test_oid_set_member_type	FP_gss_test_oid_set_member = NULL;
#endif

/*****************************************************************
Function: gssklog_gss_acquire_cred()

Description:
	do the gss_acquire_cred for the user
	
	For all the GSS functions, the vpcred_handle is a pointer
	cred_handle_plus which has the the gss_cred_id_t
******************************************************************/

static OM_uint32
gssklog_gss_acquire_cred(AUTH_METHOD * method,
		OM_uint32 *minor_status,
		void ** vpcred_handle,
		char * username,
		char * password)
{
	gss_cred_id_t  pcred_handle = 0;
	OM_uint32 major_status;
	gss_OID_set  mech_set_available = GSS_C_NO_OID_SET;
	gss_OID_set  mech_set_use = GSS_C_NO_OID_SET;
	cred_handle_plus cred_handle_p;
	int present = 0;

	/*
	 * username and password not supported in the 
	 * gssapi, so must be NULL
	 */

	if (username || password) 
		return  GSS_S_NO_CRED;

	/*
	 * we may have multiple mechs, so we need to find
	 * the one that has creds. we will prefer the Kerberos 
	 * if it present. If it fails will take any
	 * With seam at least, we need to pass the mech_used to 
	 * gss_init_sec_context too. 
	 */

	major_status = CALL_GSS(gss_indicate_mechs)(minor_status,
					  &mech_set_available);
	/* will skip any errors here, as we are trying to
	 * work with more advancced GSSAPI with multipl mechs
	 * but will use basic ones too
	 */
	if (major_status == 0 && mech_set_available)  {
		CALL_GSS(gss_test_oid_set_member)(minor_status,
				gss_mech_krb5, 
				mech_set_available,
				&present);
		if (present) {
			mech_set_use = gss_mech_krb5_set;
		}
	}

	major_status = CALL_GSS(gss_acquire_cred)(minor_status,
                      GSS_C_NO_NAME,
                       0,
                      mech_set_use,
                      GSS_C_INITIATE,
                      &pcred_handle,
                      NULL,
                      NULL);

	/* if failed and there are other mechs, try again */

	if (major_status && mech_set_use) {
		major_status = CALL_GSS(gss_acquire_cred)(minor_status,
       	              GSS_C_NO_NAME,
                       0,
                      GSS_C_NO_OID_SET,
                      GSS_C_INITIATE,
                      &pcred_handle,
                      NULL,
                      NULL);
	}

	cred_handle_p = malloc(sizeof(cred_handle_plus_desc));
	if (cred_handle_p == NULL) {
		major_status =  GSS_S_FAILURE;
		goto cleanup;
	}

	cred_handle_p->cred_handle = pcred_handle;
	if (mech_set_use) {
		cred_handle_p->mech_use = gss_mech_krb5;
	}
	*vpcred_handle = cred_handle_p;

cleanup:
	if (mech_set_available)
		CALL_GSS(gss_release_oid_set)(minor_status, &mech_set_available);

	return major_status;
}

/**********************************************************************/

static OM_uint32
gssklog_gss_init_sec_context(AUTH_METHOD * method,
			OM_uint32 * minor_status,
			void * vpcred_handle,
			void ** vpcontext_handle,
			int  peer_socket,
			char * service_princ_name) 
{
  const char this_function[] = "gssklog_gss_init_sec_context()";

  OM_uint32                        major_status = 0;
  OM_uint32                        minor_status2 = 0;
  cred_handle_plus                 cred_handle_p = (cred_handle_plus)vpcred_handle;
  gss_ctx_id_t                     context_handle = GSS_C_NO_CONTEXT;
  gss_name_t                       target_name = 0;
  gss_OID                          mech_type;
  OM_uint32                        req_flags = 0;
  OM_uint32                        time_req = 0;
  gss_channel_bindings_t           input_chan_bindings;
  gss_buffer_desc                  input_buffer_desc = {0,NULL};
  gss_buffer_desc                  output_buffer_desc = {0,NULL};
  gss_buffer_t                     output_buffer = &output_buffer_desc ;
  gss_buffer_t                     input_buffer  = &input_buffer_desc ; 
  gss_OID                          actual_mech_type;

  OM_uint32                        ret_flags = 0;
  OM_uint32                        time_rec = 0;

  int                              msg_result ;
  int                              context_established = 0 ;

    input_buffer->length = strlen(service_princ_name);
    input_buffer->value = strdup(service_princ_name);

    major_status = CALL_GSS(gss_import_name)(minor_status,
				   input_buffer,
				   my_nt_service_name,
				   &target_name) ;

    CALL_GSS(gss_release_buffer)(&minor_status2,
				      input_buffer) ;

  context_handle = GSS_C_NO_CONTEXT;
  mech_type = GSS_C_NO_OID;
  req_flags = GSS_C_MUTUAL_FLAG|GSS_C_CONF_FLAG;
  time_req = 0;
  input_chan_bindings = GSS_C_NO_CHANNEL_BINDINGS;
  ret_flags = 0 ;
  time_rec = 0 ;

  input_buffer->length = 0 ;

  mech_type = cred_handle_p->mech_use;

  while (!context_established) {

#ifdef DEBUG
    fprintf(stderr, "N %s: calling gss_init_sec_context\n", 
	    this_function) ;
	fprintf(stderr,"mech_use %p\n",mech_type);
#endif 

    major_status = CALL_GSS(gss_init_sec_context)(minor_status,
					cred_handle_p->cred_handle,
					&context_handle,
					target_name,
					mech_type,
					req_flags,
					time_req,
					input_chan_bindings,
					/* following test needed for SEAM 
					 * should be able to pass buffer with no data */
					(input_buffer->length)?input_buffer:0,
					&actual_mech_type,
					output_buffer,
					&ret_flags,
					&time_rec);

#ifdef DEBUG
    fprintf(stderr,
	    "N %s: Returned from init_sec_ctx w/token [%d]\n",
	    this_function, output_buffer->length) ;
#endif /* DEBUG */

    if (GSS_ERROR(major_status) ) {
	  goto err;
    }


    if (output_buffer->length != 0) {
		msg_result = send_message(method,
					  peer_socket,         
					  &context_handle,      
					  output_buffer,
					  0, /* 0 = no wrap */
					  0) ; /* 0 = no encrypt */

		/* gss_release_buffer(&minor_status, output_buffer); */

		if ( msg_result < 0 ) {
		  major_status = GSS_S_FAILURE;
	  	  goto err;
		}
		

#ifdef DEBUG
      fprintf(stderr, 
	      "N %s: Sent output token [%d], waiting for new token\n",
	      this_function, output_buffer->length) ;
#endif
      
	    CALL_GSS(gss_release_buffer)(&minor_status2,
					      output_buffer) ;
    }

    if (major_status & GSS_S_CONTINUE_NEEDED) {

      msg_result = receive_message(method,
				   peer_socket,
				   &context_handle,
				   input_buffer,
				   0) ; /* 0 = no unwrap */
      if ( msg_result < 0 ) {
#ifdef DEBUG
		fprintf(stderr,
			"E %s: receive_message: [%d]\n",
			this_function,
			msg_result);
#endif
		major_status = GSS_S_FAILURE;
		goto err;
      }
    
#ifdef DEBUG
      fprintf(stderr, 
	      "N %s: Received token: [%d]\n", 
	      this_function, 
	      input_buffer->length) ;
#endif

    } else {
      context_established = 1 ;
    } /* if (major_status & GSS_S_CONTINUE_NEEDED) */

  } /* while (!context_established) */


/*DEE send over the req, then read the resp */


	CALL_GSS(gss_release_name)(minor_status,
				&target_name);

	*vpcontext_handle = context_handle;
	return major_status;

err:
	CALL_GSS(gss_release_name)(minor_status,
				&target_name);

	if (context_handle) 
    CALL_GSS(gss_delete_sec_context)(minor_status, 
			   &context_handle,
			   GSS_C_NO_BUFFER) ;

	return major_status;

} 

/************************************************************/
static OM_uint32
gssklog_gss_delete_sec_context(AUTH_METHOD * method,
		OM_uint32 * minor_status,
		void ** vpcontext_handle)
{
	OM_uint32 major_status;

	major_status = CALL_GSS(gss_delete_sec_context)(minor_status,
			vpcontext_handle,
			GSS_C_NO_BUFFER);
	*vpcontext_handle = 0; 
	return major_status;

}


/************************************************************/

static OM_uint32
gssklog_gss_wrap(AUTH_METHOD * method,
	OM_uint32 *minor_status,
    void * pcontext_handle,
    int conf_req_flag,
    gss_qop_t qop_req,
     gss_buffer_t input_message_buffer,
    int * conf_state ,
    gss_buffer_t output_message_buffer)
{
   OM_uint32 maj_stat;

   maj_stat = CALL_GSS(gss_wrap)(minor_status, 
		*((gss_ctx_id_t *) pcontext_handle),
		conf_req_flag,
		qop_req,
		input_message_buffer,
		conf_state,
		output_message_buffer) ;

	return maj_stat;
}
  
/**************************************************************/

static OM_uint32
gssklog_gss_unwrap( AUTH_METHOD * method,
		OM_uint32 * minor_status,
        void * pcontext_handle,
        gss_buffer_t input_message_buffer,
        gss_buffer_t output_message_buffer,
        int * conf_state,
        gss_qop_t * qop_state)
{
	OM_uint32 maj_stat;

	maj_stat = CALL_GSS(gss_unwrap)(minor_status,
		*((gss_ctx_id_t *) pcontext_handle),
		input_message_buffer,
		output_message_buffer,
		conf_state,
		qop_state) ;
	return maj_stat;
}
/**************************************************************/
OM_uint32 
gssklog_gss_display_status(AUTH_METHOD * method,
               char * m, 
               OM_uint32 major_status, 
               OM_uint32 minor_status)
{
    OM_uint32 msg_ctx;
    OM_uint32 minor_status2;
    gss_buffer_desc msg;

    fprintf(stderr,
        "GSS-error %s: major:%8.8x minor:%8.8x\n",
        m, major_status, minor_status);

    msg_ctx = 0;
    if (major_status)
    {
        do {
            CALL_GSS(gss_display_status)(&minor_status2,
                    major_status,
                    GSS_C_GSS_CODE,
                    gss_mech_krb5 /*GSS_C_NO_OID*/,
                    &msg_ctx,
                    &msg);
            if (msg.length)
            {
                fprintf(stderr,"%.*s\n", msg.length, (char *)msg.value);
                CALL_GSS(gss_release_buffer)(&minor_status2,&msg);
            }

        } while (msg_ctx != 0);
    }

    msg_ctx = 0;
    if (1)
    {
         do {

            CALL_GSS(gss_display_status)(&minor_status2,
                        minor_status,
                        GSS_C_MECH_CODE,
                        gss_mech_krb5 /*GSS_C_NO_OID*/,
                        &msg_ctx,
                        &msg);
            if (msg.length)
            {
                fprintf(stderr,"%.*s\n", msg.length, (char *)msg.value);
                CALL_GSS(gss_release_buffer)(&minor_status2,&msg);
            }
        } while (msg_ctx != 0);
    }
    return 0;
}

/**************************************************************/
static
OM_uint32
gssklog_gss_release_buffer(AUTH_METHOD * method,
		OM_uint32 * minor_status,
		gss_buffer_t buffer)
{
  OM_uint32 major_status;
  major_status = CALL_GSS(gss_release_buffer)(minor_status,buffer);
  return major_status;
}

/**************************************************************/
static
OM_uint32
gssklog_gss_release_cred(AUTH_METHOD * method,
		OM_uint32 * minor_status,
		void ** vpcred_handle)
{
  OM_uint32 major_status;
  cred_handle_plus * pcred_handle_plus = (cred_handle_plus *)vpcred_handle;
  major_status = CALL_GSS(gss_release_cred)(minor_status, 
			&((*pcred_handle_plus)->cred_handle));

  free(*pcred_handle_plus);
  return major_status;
}

/**************************************************************/
static AUTH_METHOD gss_auth_data = {
    0,
    gssklog_gss_acquire_cred,
    gssklog_gss_init_sec_context,
    gssklog_gss_wrap,
    gssklog_gss_unwrap,
    gssklog_gss_display_status,
	gssklog_gss_release_buffer,
	gssklog_gss_release_cred, 
	gssklog_gss_delete_sec_context
    };

/*****************************************************************
******************************************************************/
AUTH_METHOD * gssklog_gss_init_method()
{
#if _WIN32
	/* On windows, we will dynamicly load the gssapi32.dll and
	 * set the entry points if any fail we return NULL 
	 */
	HMODULE gssapi_handle = NULL;
	char * failed=NULL;

	gssapi_handle = LoadLibrary("gssapi32.dll");
	if (!gssapi_handle) {
#if DEBUG
		fprintf(stderr,"Did not find gssapi32.dll, continuing\n");
#endif
		return NULL;
	}

#define FIND_ENTRY(fun) if (!(FP_##fun = (fun##_type) GetProcAddress(gssapi_handle,#fun))){failed=#fun;}
	FIND_ENTRY(gss_acquire_cred) 
	FIND_ENTRY(gss_delete_sec_context)
	FIND_ENTRY(gss_display_status)
	FIND_ENTRY(gss_import_name)
	FIND_ENTRY(gss_init_sec_context)
	FIND_ENTRY(gss_release_buffer) 
	FIND_ENTRY(gss_release_name)
	FIND_ENTRY(gss_wrap)
	FIND_ENTRY(gss_unwrap) 
	FIND_ENTRY(gss_release_cred)
	FIND_ENTRY(gss_indicate_mechs)
	FIND_ENTRY(gss_release_oid_set)
	FIND_ENTRY(gss_test_oid_set_member)
	if(failed) {
#if DEBUG
		fprintf(stderr,"Failed to find one or more gss entry points including %s\n",failed);
#endif
		return NULL;
	}
#endif
 
 return (&gss_auth_data);

}
