
/*
 * Routine to call gssklog to get an AFS token.
 * usable on machines or deamons which don't have PAM
 */

#include <stdio.h>
#include <fcntl.h>
#include <sys/wait.h>

int
gssklog_pag_klog(int flag, char ** penv) 
{
	extern char **environ;
	char ** env;
	pid_t pid;
	pid_t wpid;
	int wait_status;
	

#ifdef KRB5_HP
	char * gssklog_exec = "/krb5.hp/bin/gssklog";
#else
	char * gssklog_exec = "/krb5/bin/gssklog";
#endif
#if defined(TARGET_ARCH_AIX)
	static int (*apagaix)() = 0;
	char * apagaix_exec = "/krb5/libexec";
#endif 

	env = penv? penv: environ; 
/*
 * the -setpag does not work on AIX or SGI 
 * So get a PAG here otherwise use the -setpag option
 * of the gssklog which will set the PAG of this process
 * It also fails on Solaris 5.9
 */


/* flag = 1 says get a PAG */
	if (flag == 1) {
#if defined(TARGET_ARCH_IRIX)
    	syscall(1065,0,0,0,0,0,0); /* AFS PAG */
#elif defined(TARGET_ARCH_SOLARIS)
	    {
			char buf[256];
			long buf_len;
			buf_len = 256;
			afs_syscall = 105;
			if (sysinfo(SI_RELEASE,buf, buf_len) > 0) {
				     if (!strcmp(buf,"5.6"))  afs_syscall = 105;
				else if (!strcmp(buf,"5.7"))  afs_syscall = 73;
				else if (!strcmp(buf,"5.8"))  afs_syscall = 65;
				else if (!strcmp(buf,"5.9"))  afs_syscall = 65;
			}
		syscall(65, afs_syscall,0,0,0,0,0);
		}
#elif defined(TARGET_ARCH_AIX)
    	if (!apagaix) {
       		apagaix = load(apagaix_exec,0,0);
    	}
    	if (apagaix) {
       		(*apagaix)();
    	}
#endif
	}

	pid = fork();
	if (pid < 0) {
		return 0;
	}

	if (pid == 0) {  /* child process */

#ifdef DEBUG
		{ char ** cp;
			extern char **environ;
			fprintf(stderr,"ENV=\n");
			for(cp=environ; *cp; cp++) {
				fprintf(stderr,"   %s\n",*cp);
			}
			fprintf(stderr,"PAM_ENV=\n");
			for(cp=env; *cp; cp++) {
				fprintf(stderr,"   %s\n",*cp);
			}
		}
#endif

		/*
		 * we dont want any output to confuse the caller such as rlogind
		 * so we will close stdout and stderr
		 */

#ifndef DEBUG
		close(1);
		open("/dev/null",O_WRONLY);
		close(2);
		open("/dev/null",O_WRONLY);
#endif

		if (flag) {
		execle(gssklog_exec, "gssklog",
				"-silent",
#if !defined(TARGET_ARCH_IRIX) && !defined(TARGET_ARCH_AIX)
				"-setpag",
#endif
				 0, env);
		} else {
				execle(gssklog_exec, "gssklog",0,env);
		}
		exit (127); /* in case execl fails */
	} /* end of child */
	/* Parent, wait for child */

	wait_status = 0;
	wpid = waitpid((pid_t) pid, &wait_status, 0);

#ifdef DEBUG
	fprintf(stderr,"Waited child rc=%d\n",wait_status);
#endif

	/* alway return 0 even if failed to get token */
	return 0;
}
