/**********************************************************************

gssklogd.c

Description:
 	Given that the client can authenticate using the GSSAPI,
	issue an AFS token. 

CVS Information:

    $Source: $
    $Date: $
    $Revision: $
    $Author: $

**********************************************************************/

static char *rcsid = "$Header: $";

/**********************************************************************
                             Include header files
**********************************************************************/

#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <pwd.h>
#include <des.h>

#include <stdarg.h>
#include <sys/file.h>
#include <unistd.h>
#include <sys/wait.h>
#ifdef HAVE_PATHS_H
#include <paths.h>
#endif

#include <time.h>
#include <arpa/inet.h>
#include <syslog.h>
#include <wait.h>
#include <signal.h>

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#ifdef FREEBSD
#include <stdlib.h>
#endif /* FREEBSD */

#include <string.h>

#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN 64
#endif

#ifndef DEFAULT_PORT
#define DEFAULT_PORT 750
#endif

#include "gssklog_comm.h"
#include "gssklogd.h"

#ifdef USE_TCP_WRAPPERS
#include <tcpd.h>
int     deny_severity = LOG_WARNING;
int     allow_severity = LOG_NOTICE;
#endif

#ifndef MAXSYSLOGLEN
#define MAXSYSLOGLEN 256
#endif

/**********************************************************************
                       Define module specific variables
**********************************************************************/

char * enterprise_realm[MAX_ENTERPRISE_REALMS];

static int      debug = 0;
static int      foreground = 0;
static char *   logfile = NULL;
static char *   program_name = NULL;
static long int max_tktlifetime = 0;

/* used with GSI: */
static char *   cert_file = "/etc/grid-security/afscert.pem";
static char *   key_file  = "/etc/grid-security/afskey.pem";
static char *   cert_dir  = "/etc/grid-security/certificates";

/*used with K5: */
static char *   krb5keytab = NULL;

static char *   gridmap_file = "/etc/grid-security/afsgrid-mapfile";
static char *   afskeyfile = "/usr/afs/etc/KeyFile";

static AUTH_METHOD * method = NULL;
static gss_cred_id_t   credential_handle = GSS_C_NO_CREDENTIAL;
static gss_OID         mech_use =  GSS_C_NULL_OID;
static char *          server_name_char= NULL;

static gss_OID_desc gss_mech_krb5_desc = 
    {9, (void*)"\052\206\110\206\367\022\001\002\002"};
static gss_OID gss_mech_krb5 = &gss_mech_krb5_desc;

static gss_OID_set_desc gss_mech_krb5_set_desc =
    {1, (gss_OID) &gss_mech_krb5_desc};

static gss_OID_set gss_mech_krb5_set = &gss_mech_krb5_set_desc;

#if defined(GSS_RFC_COMPLIANT_OIDS) && (GSS_RFC_COMPLIANT_OIDS == 1) 
    /* Heimdal is not comparing OIDs, correctly, it iscomparing the pointers, */
    /* and not the value. Therefore we must use the pointer. */
#define my_nt_service_name  GSS_C_NT_HOSTBASED_SERVICE
#else
/* allocate a staic OID, as early MIT did not follow v2 RFC */

static gss_OID_desc my_nt_service_name_desc =
    {10, (void *)"\x2a\x86\x48\x86\xf7\x12\x01\x02\x01\x04"};
static gss_OID my_nt_service_name = &my_nt_service_name_desc;
#endif

void 
handle_connections(void *);

int           
daemon_me(int, int);

int 
run_acceptor(void*);

/*********************************************************************
Function:       convert_time_str()
Description:    Converts time string HH[:MM[:SS]] into long int seconds.
                Only possitive converted values are set.
Parameters:
Returns:        zero on success, -1 on error
**********************************************************************/
static int 
convert_time_str(char * time_str, long int * converted_time)
{
   char *sp;                     /* string ptr to rest of time_str */
   long int timesec = 0;
   char *str_s = time_str;             /* position to string start */
  
   timesec = 3600 * strtol (str_s, &sp, 0);               /* hours */
   if (sp == str_s) return -1;
   if (*sp == ':') {
       str_s = sp+1;                             /* skip the colon */
       timesec += 60 * strtol (str_s, &sp, 0);          /* minutes */
       if (sp == str_s) return -1;
       if (*sp == ':') {
           str_s = sp+1;                         /* skip the colon */
           timesec += strtol (str_s, &sp, 0);           /* seconds */
           if (sp == str_s) return -1;
           if (*sp) return -1;
       } else if (*sp) return -1;
   } else if (*sp) return -1;
   if (timesec > 0) {
       *converted_time = timesec;
   }
   return 0;
}

/*********************************************************************
Function:       timestamp()
Description:    
Parameters:
Returns:
**********************************************************************/
static char *
timestamp(void)
{
    time_t clock;
    struct tm *tmp;

    time(&clock);
    tmp = localtime(&clock);
    return asctime(tmp);
} /* timestamp() */

/*********************************************************************
Function:       envset()
Description:    Set an env variable
Parameters: Variable, value
Returns:
**********************************************************************/
void
envset(char *variable,char *value)
{
       char *tmp ;
       int len ;

       len = strlen(variable) + strlen(value) + 4;
       tmp = (char *) malloc(len);
       if (tmp != NULL) {
           strncpy(tmp, variable, len);
           strncat(tmp, "=", len);
           strncat(tmp, value, len);
           putenv(tmp);
           /* free( tmp ); Can not free, as it becomes part of env */
       } else {
               fprintf(stderr,"Memory error in envset");
               exit(1);
       }
}                              /* envset() */

/*********************************************************************
Function:       report messages(nargs,priority,function_name,format,... )
Description:    log errors and info
Parameters: number args, priority, function_name, format_string,...
Returns: void
**********************************************************************/
void
report (int nargs, ...)
{
    va_list args;
    int priority;
    char *function;
    char *fmt ;
    char msg[MAXSYSLOGLEN+1] ;
    va_start(args,nargs);
    priority = va_arg(args,int);

    fmt = va_arg(args,char *);
    /* build message */
    vsnprintf(msg,MAXSYSLOGLEN,fmt,args);
    if (debug ) {
       fprintf(stderr,"%s\n",msg) ;
    }
    syslog(priority,"%s",msg) ;
    va_end(args);

}  /* report() */

/*****************************************************************/

main(int argc, char **argv)
{
  
	const char this_function[] = "main()";
	int i;
	int port = DEFAULT_PORT; 
	int acceptor_socket;
	gss_buffer_desc name_buf;
	gss_name_t server_name = NULL;
	OM_uint32 major_status, minor_status;
	gss_OID_set  mech_set_available = GSS_C_NULL_OID_SET;
	gss_OID_set  mech_set_use = GSS_C_NULL_OID_SET;
	int present;
	char * lcell= NULL;
	des_cblock new_key;
	char localname[MAXHOSTNAMELEN];
	time_t now;
	int enterprise_realm_index = 0;
	char * max_tktlifetime_arg = NULL;

    if (strrchr(argv[0],'/'))
	    program_name = strrchr(argv[0],'/') + 1;
    else
   		program_name = argv[0];


#ifdef WIN32
    CRYPTO_malloc_init();
#endif

	method = gssklog_gss_init_method();

	for (i=0; i<MAX_ENTERPRISE_REALMS;i++){
		enterprise_realm[i] = NULL;
	}

	for (i=1; i<argc; i++)
	{
		if ((strcmp(argv[i], "-d") == 0))
		{
			debug = 1;
			foreground = 1;   /* Run in the forground */
		}
		else if (((strcmp(argv[i], "-p") == 0)) 
				&& (i + 1 < argc))
		{
			port = atoi(argv[i+1]);
			i++;
		}
		else if (((strcmp(argv[i], "-l") == 0))
				&& (i + 1 < argc))
		{
			logfile =  argv[i+1];
			i++;
		}
		else if (((strcmp(argv[i], "-s") == 0)) 
				&& (i + 1 < argc))
		{
			server_name_char =  argv[i+1];
			i++;
		}
		else if (((strcmp(argv[i], "-C") == 0)) 
				&& (i + 1 < argc))
		{
			cert_file =  argv[i+1];
			i++;
		}
		else if (((strcmp(argv[i], "-K") == 0)) 
				&& (i + 1 < argc))
		{
			key_file =  argv[i+1];
			i++;
		}
		else if (((strcmp(argv[i], "-D") == 0)) 
				&& (i + 1 < argc))
		{
			cert_dir =  argv[i+1];
			i++;
		}
		else if (((strcmp(argv[i], "-G") == 0)) 
				&& (i + 1 < argc))
		{
			gridmap_file =  argv[i+1];
			i++;
		}
		else if (((strcmp(argv[i], "-a") == 0)) 
				&& (i + 1 < argc))
		{
			afskeyfile =  argv[i+1];
			i++;
		}
		else if (((strcmp(argv[i], "-M") == 0))
				&& (i + 1 < argc))
		{
			max_tktlifetime_arg =  argv[i+1];
			i++;
		}
		else if (((strcmp(argv[i], "-k") == 0)) 
				&& (i + 1 < argc))
		{
			krb5keytab =  argv[i+1];
			i++;
		}
		else if (((strcmp(argv[i], "-E") == 0))
				&& (i + 1 < argc))
		{
			enterprise_realm[enterprise_realm_index] = argv[i+1];
			i++;
			enterprise_realm_index++;
			if (enterprise_realm_index >= MAX_ENTERPRISE_REALMS)
			{
				fprintf(stderr,"To many enterprise_realms\n");
				exit(2);
			}
		}
		else
		{

			fprintf(stderr, "Unknown argument %s\n", argv[i]);
			exit(1);
		}
	}

	openlog(program_name ,LOG_PID, LOG_DAEMON);

	/* set maximum allowed token lifetime */
	if (max_tktlifetime_arg != NULL) {
		int ret;
		ret = convert_time_str(max_tktlifetime_arg, &max_tktlifetime);
		if (ret != 0) {
			fprintf(stderr, 
				"translating '%s' to maximum token lifetime failed\n",
				 max_tktlifetime_arg);
			exit(1);
		}
#ifdef DEBUG
		fprintf(stderr, 
				"argument '-M %s' translated to <%d> seconds\n",
				max_tktlifetime_arg, max_tktlifetime);
#endif
	}

	if (cert_file) {
		envset("X509_USER_CERT", cert_file);
	}

	if (key_file)
	{
		envset("X509_USER_KEY",key_file);
	}

	if (cert_dir)
	{
		envset("X509_CERT_DIR",cert_dir);
	}

	if (gridmap_file)
	{
		envset("GRIDMAP",gridmap_file);
	}

	if (krb5keytab)
	{
		envset("KRB5_KTNAME",krb5keytab);
	}

  	/* get AFS cell and realm names */

	lcell = init_afs_local_cell_name();

	gethostname(localname, MAXHOSTNAMELEN);

	if (!server_name_char) {
		if  (server_name_char = (char *) malloc(strlen("gssklog@") 
				+ strlen(localname) + 1)) {
			strcpy(server_name_char,"gssklog@");
			strcat(server_name_char,localname);
		}
	}

	/* get server creds */

	if (server_name_char)
	{
		name_buf.value = server_name_char;
		name_buf.length = strlen(name_buf.value) + 1;
		major_status = gss_import_name(&minor_status, 
								&name_buf,
								my_nt_service_name,
								&server_name);
		if (major_status != GSS_S_COMPLETE)
		{
			method->auth_display_status(method,"importing name", 
							major_status, minor_status);
			exit(1);
		}
#ifdef DEBUG
	  {
		gss_buffer_desc exported_name;
		gss_OID output_name_type = NULL; 

		major_status = gss_display_name(&minor_status, server_name, 

						&exported_name, &output_name_type);
		if (major_status != GSS_S_COMPLETE)
		{
			method->auth_display_status(method,"displaying name", 
							major_status, minor_status);
			exit(1);
		}
		fprintf(stderr,"len=%d, name=%s\n",
			exported_name.length, exported_name.value);
	  }
#endif


	}
    /*
     * we may have multiple mechs, so we need to find
     * the one that has creds. we will prefer the Kerberos
     * if it present. If it fails will take any
     * With seam at least, we need to pass the mech_used to
     * gss_init_sec_context too.
     */

	major_status = gss_indicate_mechs(&minor_status,
					  &mech_set_available);
	
    /* will skip any errors here, as we are trying to
     * work with more advancced GSSAPI with multipl mechs
     * but will use basic ones too
     */
    if (major_status == 0 && mech_set_available)  {
        gss_test_oid_set_member(&minor_status,
                gss_mech_krb5,
                mech_set_available,
                &present);
        if (present) {
            mech_set_use = gss_mech_krb5_set;
			mech_use = gss_mech_krb5;
        }
    }

	major_status = gss_acquire_cred(&minor_status,
                      server_name,
                       0,
                      mech_set_use,
                      GSS_C_ACCEPT,
                      &credential_handle,
                      NULL,
                      NULL);

	/* if failed and there are other mechs, try again */
	if (major_status && mech_set_use) {
		mech_use = GSS_C_NULL_OID;
		major_status = gss_acquire_cred(&minor_status,
                      server_name,
                       0,
                      GSS_C_NULL_OID_SET,
                      GSS_C_ACCEPT,
                      &credential_handle,
                      NULL,
                      NULL);
	}
	if (major_status != GSS_S_COMPLETE) {
		method->auth_display_status(method,"acquiring credentials", 
				major_status, minor_status);
		exit(2);
	}
	gss_release_name(&minor_status, &server_name);

  /* Make sure we have a random number generator */
  /* since we will be generating session keys for tokens */
  /* The AFS des routines will seed this the first time */
  /* But if using K5 or OPenSSL they will not. */
  /* We don't need to keep the key */

  /* THIS NEEDS WORK !!! */

	time(&now);

#ifdef USE_OPENSSL_DES
	des_random_seed((des_cblock *)&time);
#endif
#ifdef USE_KRB5_DES
	des_init_random_number_generator((des_cblock *)&time);
#endif
	des_random_key(new_key);



   /*
    * Do the standard daemon stuff.
    */
    if ( ! debug ) {
       if ( daemon_me(0,0) ) {
           exit (1) ;
       }
    }

	/* open communication link and start listening. */

    openlog(program_name, LOG_PID, LOG_DAEMON);
    /*
     * Set up default signal handlers.
     */
    setup_signals();

    acceptor_socket=setup_connection(port);
    handle_connections((void *) acceptor_socket);

} /* main */


/*****************************************************************
Function: doexit

Description:
       log failing message to syslog.
******************************************************************/
void
doexit(int sig ) {
    report(3,LOG_ERR,"Dieing with signal %d",sig );
    closelog();
    exit(sig) ;
}

/*****************************************************************
Function: do_child

Description:
       Start child deamon and wait. Report if a problem. 
******************************************************************/
void
do_child(int sig) {
    pid_t pid ;
    int pid_stat ;
    /* Get them all while we're here */
    while (  pid = waitpid(-1,&pid_stat,WNOHANG) > 0 ) {
       if ( WIFEXITED(pid_stat) ) {
		  if ( WEXITSTATUS(pid_stat) ) {       /* exit(0) is normal */ 
            report(4,LOG_ERR,
                  "Child %d exited with status %d",
                  pid,WEXITSTATUS(pid_stat));
		  }
       } else {
           report(3,LOG_INFO,"Abnormal child termination %d",pid_stat);
       }
    }
}

/*****************************************************************
Function: display_status_report

Description:
       display the gss errors via report mech.
******************************************************************/
int
setup_signals()
{
    struct sigaction new_act;

    new_act.sa_handler = doexit;
    sigemptyset(&new_act.sa_mask);
    sigaction(SIGINT, &new_act, 0);
    sigaction(SIGTERM, &new_act, 0);
    sigaction(SIGHUP, &new_act, 0);
    sigaction(SIGQUIT, &new_act, 0);
    sigaction(SIGALRM, &new_act, 0);
    new_act.sa_handler = SIG_IGN;
    sigaction(SIGPIPE, &new_act, 0);
    new_act.sa_handler = do_child;
    sigaction(SIGCHLD, &new_act, 0);

}


/*****************************************************************
Function: display_status_report

Description:
       display the gss errors via report mech.
******************************************************************/

OM_uint32
display_status_report(char *m, OM_uint32 major_status, OM_uint32 minor_status)
{
    OM_uint32       msg_ctx;
    OM_uint32       minor_status2;
    gss_buffer_desc msg;

    report(5,LOG_ERR,
           "GSS-error %s: major_status:%8.8x minor_status:%8.8x\n",
           m, major_status, minor_status);

    msg_ctx = 0;
    if (major_status) {
       do {
           gss_display_status(&minor_status2,
                              major_status,
                              GSS_C_GSS_CODE, mech_use, &msg_ctx,
                              &msg);
           if (msg.length) {
               report(4,LOG_ERR, "%.*s\n", msg.length, (char *) msg.value);
               gss_release_buffer(&minor_status2, &msg);
           }
       }
       while (msg_ctx != 0);
    }

    msg_ctx = 0;
    if (1) {
       do {
           gss_display_status(&minor_status2,
                              minor_status,
                              GSS_C_MECH_CODE,
                              mech_use, &msg_ctx, &msg);
           if (msg.length) {
               report(4,LOG_ERR,"%.*s\n", msg.length, (char *) msg.value);
               gss_release_buffer(&minor_status2, &msg);
           }
       }
       while (msg_ctx != 0);
    }
    return 0;
}

/*****************************************************************/
void handle_connections(void *opq)
{
  char this_function[] = "handle_connections";
  int my_socket = (int) opq;
  int s;
  int pid = 0;
  des_cblock new_key;

  while(1) {
    s = establish_connection(my_socket);
    if (s >= 0) {

#ifdef DEBUG
      fprintf(stderr,
	      "N %s: got connection, s = %d\n", this_function, s);
#endif

	  if (!foreground)
	    pid = fork();
	  if (pid < 0) {
		report(4,LOG_ALERT, "Fork error children" ) ;
		exit (4);
	  }
		
	  if (pid == 0) {
        run_acceptor((void *)s);
		if (!foreground) {
		  close(my_socket);
		  closelog();
		  close(s);
	      exit(0); /* end child process */
        }
	  }
     
	  wait(0); /* give up control to give child a chance? */
      close(s);

	/*
	 * Make sure next random number starts with a new seed
	 * we don't want to call the next child with the same seed
	 * as this has the possibility of generating the same key.
	 * So depending on the internals of the random_key routines 
	 * this new_key could be the previous session key or not.
	 * in any case the next child will start with a new seed.
	 */
	
	des_random_key(new_key);

#ifdef DEBUG
      fprintf(stderr,
	      "N %s: Listening for next.\n", this_function) ;
#endif

    } else {
		report(4,LOG_ALERT,
               "W %s: didn't get connection, s = %d",
                this_function, s);
    }
  }
} /* handle_connections () */

/*****************************************************************/
int run_acceptor (void *opq) 
{
  char this_function[] = "run_acceptor";


  int                            acceptor_socket;


  gss_buffer_desc                input_token_desc = {0, NULL};
  gss_buffer_desc                output_token_desc = {0,NULL};
  gss_buffer_desc                tmp_buffer_desc = {0, NULL};
  gss_buffer_t                   input_token = &input_token_desc;
  gss_buffer_t                   output_token = &output_token_desc;
  gss_buffer_t                   tmp_buffer= &tmp_buffer_desc;

  gss_ctx_id_t                   context_handle;
  gss_name_t                     client_name = NULL;
  char *                         client_name_char = NULL;
 
  int                            msg_result ;
  OM_uint32                      major_status = 0, minor_status=0 ;
  OM_uint32                      minor_status2;
  OM_uint32                      glifetime = GSS_C_INDEFINITE;
  OM_uint32						 ret_flags = 0;

  unsigned char * req;
  unsigned char * resp;
  int reqlen;
  int resplen;
  int reqcode;

  struct hostent     *hp;
  struct sockaddr_in peer;
  size_t peerlen;
  char * peernum;
  char * peerchar;
  char logtime[64];
  char logbuf[1024] = "No Log Entry";
  char rname[2*MAXKTCNAMELEN+1];

#ifdef USE_TCP_WRAPPERS
	struct request_info wrapper_request;
#endif

#ifdef DEBUG
  fprintf(stderr,"N %s: initiated on %d\n", this_function, (int) opq) ;
#endif

  acceptor_socket = (int) opq;
  context_handle = GSS_C_NO_CONTEXT ;

  /* log connection */
  
  peernum = "";
  peerchar = "";
  peerlen = sizeof(peer);
  if (getpeername(acceptor_socket, 
				(struct sockaddr *) &peer, 
				&peerlen) == 0) {
	if (peer.sin_family == AF_INET) {
		peernum = inet_ntoa(peer.sin_addr);
        hp = gethostbyaddr((char *) &(peer.sin_addr.s_addr), 
							4, AF_INET);
		if (hp) {
		  peerchar = (*hp).h_name;
		}
	} 
  }

  sprintf(logtime,"%.*s",
			19,  timestamp());

#ifdef USE_TCP_WRAPPERS

	request_init(&wrapper_request, RQ_DAEMON, program_name, 
							       RQ_FILE, acceptor_socket, 0);
	fromhost(&wrapper_request);
	if (!hosts_access(&wrapper_request)) {
        sprintf(logbuf, "%.*s[%.*s] Rejected by wrapper\n",
            200, peerchar,
            16, peernum);
        syslog(LOG_ERR, logbuf);
		goto err2;
	}
#endif

  /* do authentication */

  do {

    msg_result = receive_message(method,
				 acceptor_socket,
				 &context_handle,
				 input_token,
				 0) ;
    if ( msg_result < 0 ) {
       major_status = 
            GSS_S_DEFECTIVE_TOKEN | GSS_S_CALL_INACCESSIBLE_READ;
       break;
    }	
	
#ifdef DEBUG
    fprintf(stderr,
	    "N %s: calling gss_accept_sec_context\n", 
	    this_function) ;
#endif

    major_status = gss_accept_sec_context(&minor_status,
					  &context_handle,
					  credential_handle,
					  input_token,
					  GSS_C_NO_CHANNEL_BINDINGS,
					  &client_name,
					  NULL,
					  output_token,
					  NULL,
					  &ret_flags,
					  NULL);

    if (output_token->length != 0) {
	
#ifdef DEBUG
        fprintf(stderr,
		"N %s: sending output token: [%d]\n", 
		this_function, output_token->length) ;
#endif /* DEBUG */

      msg_result = send_message(method,
				  acceptor_socket,
				  &context_handle,
				  output_token,
				  0,
				  0) ;
      if ( msg_result < 0 ) {
        major_status = 
            GSS_S_DEFECTIVE_TOKEN | GSS_S_CALL_INACCESSIBLE_WRITE;
	  }
	  gss_release_buffer(&minor_status2,
					      output_token) ;
    } 
	
    if (GSS_ERROR(major_status)) { 
      if (context_handle != GSS_C_NO_CONTEXT) 
	     gss_delete_sec_context(&minor_status2,
						  &context_handle,
						  GSS_C_NO_BUFFER);
	  break ;
    }
    if (input_token->length >0) {
      free(input_token->value);
      input_token->length = 0;

	}

  } while (major_status & GSS_S_CONTINUE_NEEDED) ;

  if (input_token->length >0) {
    free(input_token->value); 
    input_token->length = 0;
  }

  if (major_status != GSS_S_COMPLETE) {
	display_status_report("accepting credentials", 
                   major_status, minor_status);
	goto err;
  }

  if (client_name) 
  {
     major_status = gss_display_name(&minor_status2,
                           client_name,
                           tmp_buffer,
                           NULL);
        if (major_status == GSS_S_COMPLETE) {
		  char * cp;

          cp = (char *)malloc(tmp_buffer->length+1);
          if (cp) {
            memcpy(cp, tmp_buffer->value, tmp_buffer->length);
            cp[tmp_buffer->length] = '\0';
            client_name_char = cp;
          } else {
            major_status = GSS_S_FAILURE;
          }
      }
      gss_release_buffer(&minor_status2, tmp_buffer);
  } else {
	report(2,LOG_ERR,"No client name returned\n");
	goto err;
  }




#ifdef DEBUG
    fprintf(stderr,"N %s: security context accepted\n", this_function) ;
	fprintf(stderr,"N: client_name:%s\n",client_name_char);
#endif

/*DEE need to receive the req, then send token */

    /* read in the request */
  msg_result = receive_message(method,
				   acceptor_socket,
                   &context_handle,
                   input_token,
                   1) ; /* 1 = unwrap */
  if ( msg_result < 0 ) {
        report(3,LOG_ERR, "E %s: Could not receive message. result=%d\n",
                 this_function, msg_result) ;
     goto err;
  }

  req = input_token->value;
  reqlen = input_token->length;
  if (!req) {
	goto err;
  }

  /* need to get the time, from gssapi, or some other method. */
	major_status = gss_inquire_context(&minor_status,
						context_handle,
						NULL,
						NULL,
						&glifetime,
						NULL,
						NULL,
						NULL,
						NULL);

  if (major_status != GSS_S_COMPLETE) {
	display_status_report("getting lifetime from context", 
                   major_status, minor_status);
	goto err;
  }

/* Need to check that the glifetime is valid.
 */

 reqcode = process_req(req, reqlen, 
			htonl(peer.sin_addr.s_addr),
			afskeyfile,
			client_name_char, 
			rname,  /* requested name if any */
			glifetime, max_tktlifetime,
			&resp, &resplen);

  output_token->length = resplen;
  output_token->value = resp;

  msg_result = send_message(method,
				  acceptor_socket,
                  &context_handle,
                  output_token,
                  1, /* 1 =  wrap */
                  1) ; /* 1 = encrypt */

  gss_release_buffer(&minor_status, output_token);
  if ( msg_result < 0 ) {
		report(3,LOG_ERR,"E %s: Could not send message.\n", this_function);
     goto err;
  }

  if (reqcode == 0) {
		sprintf(logbuf, "%.*s[%.*s] AFS token for %.*s to %.*s\n",
			 200, peerchar,
			 16, peernum,
			 200, rname, 
			 200,client_name_char);
		syslog(LOG_NOTICE, logbuf);
  } else {
		sprintf(logbuf, "%.*s[%.*s] AFS token failed for %.*s\n",
			200, peerchar,
			16, peernum,
			200, client_name_char);
		syslog(LOG_ERR, logbuf);
  }
	
#ifdef DEBUG
    fprintf(stderr,"%s - %s\n", logtime, logbuf);
    fprintf(stderr,"N %s: security context being shutdown\n", this_function) ;
#endif  
  
	goto ok;

err:
	sprintf(logbuf, "%.*s[%.*s] FAILED for other reasons",
		200, peerchar,
		16, peernum);
	report(3,LOG_ERR,"%s", logbuf);

err2:
#ifdef DEBUG 
	fprintf(stderr,"%s - %s\n", logtime, logbuf);
#endif

ok:
	if (client_name_char)
		free(client_name_char);

      gss_delete_sec_context(&minor_status2,
						&context_handle,
						GSS_C_NO_BUFFER);


  return 0 ;
} /* run_acceptor() */

/**************************************************************/

/*
 * Do the standard deamon things.
 */

/*
 * Copyright (c) 1990 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *     This product includes software developed by the University of
 *     California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef _PATH_DEVNULL
#define _PATH_DEVNULL "/dev/null"
#endif

int
daemon_me(nochdir, noclose)
       int nochdir, noclose;
{
       int cpid;

       if ((cpid = fork()) == -1)
               return (-1);
       if (cpid)
               exit(0);
#ifdef HAVE_SETSID
       (void) setsid();
#else
#ifndef TIOCNOTTY
       setpgrp();
#else
       {
               int n;

               /*
                * The open below may hang on pseudo ttys if the person
                * who starts named logs out before this point.  Thus,
                * the need for the timer.
                */
               alarm(120);
               n = open("/dev/tty", O_RDWR);
               alarm(0);
               if (n > 0) {
                       (void) ioctl(n, TIOCNOTTY, (char *)NULL);
                       (void) close(n);
               }
       }
#endif
#endif
       if (!nochdir)
               (void) chdir("/");
       if (!noclose) {
               int devnull = open(_PATH_DEVNULL, O_RDWR, 0);

               if (devnull != -1) {
                       (void) dup2(devnull, 0);
                       (void) dup2(devnull, 1);
                       (void) dup2(devnull, 2);
                       if (devnull > 2)
                               (void) close(devnull);
               }
       }
       return (0);
}
