/*********************************************************************

 gssklogd_afs.c 

Description:

	client utility routiones need to process the AFS tokesn
	much of this taken from openAFS kauth/krb_udp.c
	which has this copyright:
*/

/* Copyright (C) 1990 Transarc Corporation - All rights reserved */

/*
 * (C) COPYRIGHT IBM CORPORATION 1988, 1989
 * LICENSED MATERIALS - PROPERTY OF IBM
 */

/* 
CVS Information:

    $Source: $
    $Date: $
    $Revision: $
    $Author: $

********************************************************************/

static char *rcsid = "$Header: $";

/**********************************************************************
                             Include header files
**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#include <fcntl.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include <sys/time.h>
#include <arpa/inet.h>

#include <des.h>

#include <afs/param.h>
#include <afs/stds.h>
#include <afs/kautils.h>
#include <afs/keys.h>

#include "gssklogd.h"
#include "gssklog_comm.h"
#include "gssklog_afs.h"
#include "lifetimes.h"


struct packet {
    int   len;
    struct sockaddr_in from;
    int   byteOrder;
    char *name;
    char *inst;
    char *realm;
    Date  time;
    char *rest;				/* remaining bytes of packet */
    char  data[MAX_PKT_LEN];
};

static int assemble_athena_ticket();
unsigned char time_to_life ();
/* extern gss_OID gss_nt_service_name; */

static struct afsconf_dir *configdir = NULL; /* AFS config */
static char  lcell[MAXKTCREALMLEN];
static char  lrealm[MAXKTCREALMLEN];
extern char * enterprise_realm[MAX_ENTERPRISE_REALMS];

/******************************************************************/
char *
init_afs_local_cell_name()
{
  	/* get AFS cell and realm names */

	if (!(configdir = afsconf_Open(AFSDIR_CLIENT_ETC_DIRPATH))) {
		fprintf(stderr,"Unable to open AFS config file\n");
		exit (5);
	}

	if (afsconf_GetLocalCell(configdir, lcell, sizeof(lcell))) {
		fprintf(stderr,"Can not get local cell name\n");
		exit(5);
	}
	afsconf_Close(configdir);
	ucstring(lrealm, lcell, MAXKTCREALMLEN);
	return lcell;
}
/***************************************************************/
/*
 * Get highest key for this AFS cell from copy of the KeyFile
 */

int get_afs_key_from_file( char * afskeyfile,
			afs_int32  *afskvno,
  			struct ktc_encryptionKey *afskey)
{
    int fd, i;

    struct afsconf_keys afskeys;

    /* If we have not opened the file, or if the call name if
     * different read the keyfile if it exists. If not return 0.
     * This should really use a link list.
     */

  if (((fd = open(afskeyfile, O_RDONLY)) < 0) ||
       (read(fd,&afskeys,sizeof(afskeys)) != sizeof(afskeys)))
      return (1);
  close(fd);

  i = ntohl(afskeys.nkeys) - 1;
  *afskvno = ntohl(afskeys.key[i].kvno);
  memcpy(afskey, afskeys.key[i].key, sizeof(afskeys.key[i].key));

  return(0);
}
/***************************************************************/

static 
afs_int32 create_cipher (cipher, cipherLen, sessionKey, sname, sinst,
			   start, end, kvno, ticket, ticketLen, key)
  char *cipher;
  int  *cipherLen;
  struct ktc_encryptionKey *sessionKey;
  char *sname;
  char *sinst;
  Date  start, end;
  afs_int32  kvno;
  char *ticket;
  int   ticketLen;
  struct ktc_encryptionKey *key;
{   char *answer;
    int   slen;
    unsigned char life = time_to_life (start, end);
    int   len;
    des_key_schedule schedule;
    afs_int32  code;

    answer = cipher;
    len = sizeof(*sessionKey) + strlen(sname) + strlen(sinst) +
	strlen(lrealm) + 3/*nulls*/ + 3 + ticketLen + sizeof(Date);
    if (len > *cipherLen) return KAANSWERTOOLONG;
    if (ticketLen > 255) return KAANSWERTOOLONG;
    if (kvno > 255) return KAANSWERTOOLONG;

    memcpy ( answer, sessionKey, sizeof(*sessionKey));
    answer += sizeof(*sessionKey);
    putstr (answer, sname);
    putstr (answer, sinst);
    putstr (answer, lrealm);
    *answer++ = life;
    *answer++ = (unsigned char) kvno;
    *answer++ = (unsigned char) ticketLen;
    memcpy (answer, ticket, ticketLen);
    answer += ticketLen;
    putint (answer, start);


#if 0
	/* Since we will be using GSSAPI wrap we dont need to
	 * encrypt in the user key.
	 */
    if (code = des_key_sched (key, schedule))
	fprintf (stderr,"In KAAuthenticate: key_sched returned %d\n", code);
    des_pcbc_encrypt (cipher, cipher, len, schedule, key, ENCRYPT);
    *cipherLen = round_up_to_ebs (len);
#endif
	*cipherLen = len;

    return 0;
}

static 
afs_int32 create_reply (
  char * ans,
  int  * anslen, 
  char *name,
  char *inst,
  char * cell,
  Date  startTime,
  Date  endTime,
  afs_int32  kvno,
  char * cipher,
  int   cipherLen)
{   
	char *answer = ans;
    int   slen;
	int   newanslen;

    newanslen =2 + strlen(name) + strlen(inst) + 3/*nulls*/ +
	strlen(cell) +
	sizeof(afs_int32) + 1/*ntkts*/ + sizeof(afs_int32) + 1/*kvno*/ +
	sizeof(short) + cipherLen;
    if (newanslen > *anslen) return KAANSWERTOOLONG;
    if (kvno > 255) return KAANSWERTOOLONG;
	
    *answer++ = (unsigned char) KRB_PROT_VERSION;
    *answer++ = (unsigned char) AUTH_MSG_KDC_REPLY;
    /* always send claiming network byte order */
    putstr (answer, name);
    putstr (answer, inst);
    putstr (answer, cell); /*DEE*/ 
    putint (answer, startTime);
    *answer++ = 1;			/* undocumented number of tickets! */
    putint (answer, endTime);
    *answer++ = (unsigned char) kvno;
    {   short w = (short) cipherLen;
	w = htons(w);
	memcpy (answer, &w, sizeof(short));
	answer += sizeof(short);
    }
    memcpy ( answer, cipher, cipherLen);
	*anslen = newanslen;
    return 0;
}

static err_packet (char * ans, 
	int *  anslen,
	char * name,
	char * inst,
	char * realm,
	afs_int32  code,
	char *reason)
{ 
	char *answer = ans;
    int   newanslen;
	int slen;
    char  buf[256];
	Date  time;

    if (reason == 0) reason = "";
    else if (strlen(reason) + 20 > sizeof(buf)) reason = "reason too long";
    sprintf (buf, "code = %d: %s", code, reason);

    newanslen = 2 + strlen(name) + strlen(inst) + strlen(realm) +
	3/* nulls */ + 2*sizeof(afs_int32) + strlen (buf) + 1;
    if (newanslen > *anslen) {
	fprintf (stderr,"Answer packet too long\n");
	return;
    }

    *answer++ = (unsigned char) KRB_PROT_VERSION;
    *answer++ = (unsigned char) AUTH_MSG_ERR_REPLY;
    /* always send claiming network byte order */
    putstr (answer, name);
    putstr (answer, inst);
    putstr (answer, realm);
	time = 0;
    putint (answer, time);
    if ((code < 0) || (code > KERB_ERR_MAXIMUM)) {
	/* It could be because of kauth errors so we should return something instead of success!! */
	code = KDC_GEN_ERR;
    }
    putint (answer, code);
    putstr (answer, buf);
	*anslen = newanslen;
}

/*******************************************************************/
static int 
parse_req(unsigned char * req,
            int reqlen,
			struct ktc_principal  * client,
			struct ktc_principal  * server,
            afs_int32 * lifetime)
{
	unsigned char * pointer; 
	unsigned char * endpointer;
	unsigned char version, auth_msg_type;
	int   byteOrder;
	int slen;
	afs_int32 code;
	
	if (reqlen = 11 && !strncmp(req,"The Request",11)) {
		strcpy(client->name, "princ_from_cert");
		client->instance[0] = '\0';
		strcpy(client->cell, lrealm); 
		strcpy(server->name, "afs");
		server->instance[0] = '\0';
		strcpy(server->cell,lrealm);
		*lifetime = 0;
	} else {
		pointer = req;
		endpointer = pointer + reqlen;
		version = *pointer++;

		if (version != KRB_PROT_VERSION) {
			return 1;
		}
		byteOrder = *pointer++;
		auth_msg_type = byteOrder & ~1;
		if (auth_msg_type != AUTH_MSG_KDC_REQUEST) {
			return 2;
		}
		pointer++; /* skip my version, as it is 0 */
		getstr(pointer, client->name);
		getstr(pointer, client->instance);
		getstr(pointer, client->cell);
		
		getstr(pointer, server->name);
		getstr(pointer, server->instance);
		getstr(pointer, server->cell);
		getint(pointer, *lifetime);
	}
	return 0;
}

/*******************************************************************/
int
process_req(unsigned char * req,
			int reqlen,
			afs_int32 peer,
			char * afskeyfile,
			char * subject_name,
			char * rname,
			int glifetime,
			long int maxtktlifetime,
			unsigned char ** resp,
			size_t * resplen)
{
  afs_int32  afskvno = 0;
  struct ktc_encryptionKey afskey;
  Date  now = time(0);
  Date  startTime;
  Date  endTime;
  afs_int32  code;
  char * reason = "";
  char * gname = NULL;
  char * krealm = NULL;
  char gname2[2*MAXKTCNAMELEN+1];

  char  ticket[MAXKTCTICKETLEN];
  int   ticketLen;
  struct ktc_encryptionKey sessionKey;

  char  cipher[2*MAXKTCTICKETLEN];
  int   cipherLen;
  char  ans[3*MAXKTCTICKETLEN];
  int   anslen;
  int i;
  struct ktc_principal client;
  struct ktc_principal server;

  afs_int32 rlifetime;


  if ( code = parse_req(req, reqlen,
			&client, &server,
			&rlifetime)) {
		reason = "Invalid request";
		goto abort;
  }

  if (strcmp(server.name,"afs")) {
		reason = "Only support afs tickets";
		goto abort;
  }
  if (strcmp(server.cell,lcell)) {
		reason = "Only support local server cell";
		goto abort;
  }
  if (strcmp(client.cell,lcell)) {
		reason = "Only support local client cell";
		goto abort;
  }

#ifdef DEBUG
	fprintf(stderr,"Looking in gridmap for %s : %s\n",
     subject_name, client.name);
#endif

  if (!strcmp(client.name,"princ_from_cert")) {	
	code = globus_gss_assist_gridmap(subject_name, &gname);
	if (gname) {
		strcpy(client.name,gname);
	}
  } else {
	strcpy(gname2,client.name);
	if ((client.instance[0])) {
			strcat(gname2,".");
			strcat(gname2,client.instance);
	}
	code = globus_gss_assist_userok(subject_name, gname2);
  }

  if (code) {
	/*
	 * If we have an enterprise naming convension,
	 * which we do at ANL, using Unique Name,
	 * we can assume the same name in a principal matches
	 * the same name in AFS.
	 * Therefore if the name is a Kerberos principal name
	 * with a single component, in one of the enterprise
	 * realms, we can map the name in the principal to 
	 * to the AFS username, thus avoiding having to have 
	 * an entry in the map file. 
	 */
	if (enterprise_realm[0] && !client.instance[0]) {
		if (!strchr(subject_name,'/')) { /* not GSI, and not compound */
			if (krealm = strchr(subject_name,'@')) {
				krealm++;
				i = 0;
				while (enterprise_realm[i] 
					&& strcmp(enterprise_realm[i], krealm)) {
					i++;
				}

				if (enterprise_realm[i]) {
					gname = malloc(krealm-subject_name);
					if (gname) {
						strncpy(gname, subject_name,
						krealm-subject_name-1);
						gname[krealm-subject_name-1] = '\0';
					}
				}
			}
		}
  	}
	if (gname) {

		if (!strcmp(client.name,"princ_from_cert")) {
				strcpy(client.name,gname);
		} 
		if (strcmp(gname, client.name)) {
			reason = "User and credential do not agree";
			goto abort;
		}
	} else {
		reason = "Unable to map to AFS user";
		goto abort;
  	}
  }
	
#ifdef DEBUG
	fprintf(stderr,"cname=%s cell=%s rl=%d gl=%d\n",
		client.name, client.cell, rlifetime, glifetime);
#endif

  code = des_random_key (&sessionKey);

  /* 
   * Define a lifetime. If glifetime not known, default to 
   * requested lifetime. If requested lifetime is  zero, default 
   * to credential lifetime. If both zero, use 8 hours. 
   */
  if (glifetime == GSS_C_INDEFINITE ) glifetime = rlifetime;
  if (rlifetime == 0) rlifetime = glifetime;
  if (rlifetime == 0 && glifetime == 0) rlifetime = glifetime = 8*60*60;

  if (rlifetime> MAXTKTLIFETIME) rlifetime = MAXTKTLIFETIME;
  if (glifetime> MAXTKTLIFETIME) glifetime = MAXTKTLIFETIME; 
  /* If maxtktlifetime argument was set for server, 
	 limit to that value as well */
  if (maxtktlifetime > 0) {
	if (rlifetime > maxtktlifetime) rlifetime = maxtktlifetime;
	if (glifetime > maxtktlifetime) glifetime = maxtktlifetime;
  }


  startTime = now;
  endTime = startTime + ((glifetime<=rlifetime)?glifetime:rlifetime);

#ifdef DEBUG
	fprintf(stderr,"Times:startTime=%d,endTime=%d,glife=%d,rlife=%d\n",
		startTime,endTime,glifetime,rlifetime);
#endif
    if (code = get_afs_key_from_file(afskeyfile, &afskvno, &afskey)) {
		reason = "Internal error: get_afs_key_from_file"; 
		goto abort;
	}

	ticketLen = sizeof(ticket);
	if (code = tkt_MakeTicket(ticket, &ticketLen, &afskey,
			client.name, client.instance, lrealm,
			startTime, endTime, &sessionKey,
			0,
			server.name, server.instance)) {
fprintf(stderr,"FAILED:%s %s \n",
		client.name, client.instance);
				reason = "Internal error: tkt_MakeTicket";
				goto abort;
	}

	cipherLen = sizeof(cipher);
	if (code = create_cipher(cipher, &cipherLen, &sessionKey, 
					server.name, server.instance,
					startTime, endTime, 
					afskvno, ticket, ticketLen, 
					NULL)) {
			reason = "Internal error: create_cipher";
			goto abort;
	}
	anslen = sizeof(ans);
	if (code = create_reply(ans, &anslen, 
					  client.name, client.instance, lcell,
 					  startTime, endTime, 0,
                      cipher, cipherLen)) {
			 reason = "Internal error: create_reply";
			goto abort;
	}
					
				

	*resp = malloc(anslen);
	*resplen = anslen;
	memcpy( *resp, ans, *resplen);

	if (rname) {
		strcpy(rname,client.name);
		if (client.instance[0]) {
			strcat(rname,".");
			strcat(rname,client.instance);
		}
	}

	return 0;
abort:
	anslen = sizeof(ans);
	err_packet(ans, &anslen, client.name, client.instance, 
				lrealm, code, reason);
	*resp = malloc(anslen);
    *resplen = anslen;
   memcpy (*resp, ans, *resplen);
	
#ifdef DEBUG
	fprintf (stderr,"FAILED code=%d reason=%s\n",
			code, reason?reason:"");
#endif
	return code;

}

/******************************************************************/
afs_int32 ktohl_internal (char flags, afs_int32 l)
{
    if (flags & 1) {
	unsigned char *lp = (unsigned char *)&l;
	afs_int32	       hl;
	hl = *lp + (*(lp+1) << 8) + (*(lp+2) << 16) + (*(lp+3) << 24);
	return hl;
    }
    return ntohl(l);
}

/******************************************************************/
/* time_to_life - takes start and end times for the ticket and returns a
 * Kerberos standard lifetime char possibily using the tkt_lifetimes table for
 * lifetimes above 127*5minutes.  First, the special case of (end ==
 * 0xffffffff) is handled to mean no expiration.  Then negative lifetimes and
 * those greater than the maximum ticket lifetime are rejected.  Then lifetimes
 * less than the first table entry are handled by rounding the requested
 * lifetime *up* to the next 5 minute interval.  The final step is to search
 * the table for the smallest entry *greater than or equal* to the requested
 * entry.  The actual code is prepared to handle the case where the table is
 * unordered but that it an unnecessary frill. */

unsigned char time_to_life (start, end)
  afs_uint32 start;
  afs_uint32 end;
{   int lifetime = end-start;
    int best, best_i;
    int i;

    if (end == NEVERDATE) return TKTLIFENOEXPIRE;
    if ((lifetime > MAXKTCTICKETLIFETIME) || (lifetime <= 0)) return 0;
    if (lifetime < tkt_lifetimes[0]) return (lifetime + 5*60-1) / (5*60);
    best_i = -1;
    best = MAXKTCTICKETLIFETIME;
    for (i=0; i<TKTLIFENUMFIXED; i++)
	if (tkt_lifetimes[i] >= lifetime) {
	    int diff = tkt_lifetimes[i]-lifetime;
	    if (diff < best) {
		best = diff;
		best_i = i;
	    }}
    if (best_i < 0) return 0;
    return best_i+TKTLIFEMINFIXED;
}


/******************************************************************/
/* This makes a Kerberos ticket */

int tkt_MakeTicket (ticket, ticketLen, key, name, inst, cell,
		    start, end, sessionKey, host, sname, sinst)
  char		*ticket;		/* ticket is constructed here */
  int		*ticketLen;		/* output length of finished ticket */
  struct ktc_encryptionKey *key;	/* key ticket should be sealed with */
  char		*name;			/* user of this ticket */
  char		*inst;
  char		*cell;			/* cell of authentication */
  afs_uint32	 start,end;		/* life of ticket */
  struct ktc_encryptionKey *sessionKey;	/* session key invented for ticket */
  afs_uint32	 host;			/* caller's host address */
  char		*sname;			/* server */
  char		*sinst;
{   int		 code;
    Key_schedule schedule;

    *ticketLen = 0;			/* in case we return early */
    code = assemble_athena_ticket (ticket, ticketLen, name, inst, cell,
				   host, sessionKey, start, end, sname, sinst);
    *ticketLen = round_up_to_ebs(*ticketLen); /* round up */
    if (code) return -1;

    /* encrypt ticket */
    if (code = key_sched (key, schedule)) {
	fprintf (stderr,"In tkt_MakeTicket: key_sched returned %d\n", code);
	return RXKADBADKEY;
    }
    pcbc_encrypt (ticket, ticket, *ticketLen, schedule, key, ENCRYPT);
    return 0;
}

#undef putstr
#undef putint

#define putstr(name,min) \
    slen = strlen(name); \
    if ((slen < min) || (slen >= MAXKTCNAMELEN)) return -1; \
    strcpy (ticket, name); \
    ticket += slen+1
#define putint(num) num = htonl(num);\
		    memcpy (ticket, &num, sizeof(num));\
		    ticket += sizeof(num)

/******************************************************************/
static int 
assemble_athena_ticket (ticket, ticketLen, name, inst, realm,
				    host, sessionKey, start, end, sname, sinst)
  char *ticket;
  int  *ticketLen;
  char *name;
  char *inst;
  char *realm;
  afs_int32  host;
  struct ktc_encryptionKey *sessionKey;
  afs_uint32 start;
  afs_uint32 end;
  char *sname;
  char *sinst;
{   char *ticketBeg = ticket;
    int	  slen;
    unsigned char  life;

    *ticket++ = 0;			/* flags, always send net-byte-order */
    putstr (name, 1);
    putstr (inst, 0);
    putstr (realm, 0);
    putint (host);

    memcpy (ticket, sessionKey, sizeof(struct ktc_encryptionKey));
    ticket += sizeof(struct ktc_encryptionKey);

    life = time_to_life (start, end);
    if (life == 0) return -1;
    *ticket++ = life;

    putint (start);
    putstr (sname, 1);
    putstr (sinst, 0);

    *ticketLen = ticket - ticketBeg;
    return 0;
}

